#!/usr/bin/env python3

import stepwise, appcli, autoprop
from appcli import Key, DocoptConfig
from stepwise import StepwiseConfig, PresetConfig, pl, ul
from stepwise_mol_bio import Main, merge_dicts, format_min, format_sec
from inform import plural
from collections import defaultdict

@autoprop
class Stain(Main):
    """\
Stain a gel.

Usage:
    stain <preset> [-t <min>] [-i <cmd> | -I]

<%! from stepwise_mol_bio import hanging_indent %>\
Arguments:
    <preset>
        The default parameters to use.  The following presets are available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

Options:
    -t --time <min>
        How long to incubate the gel in the stain, in minutes.

    -i --imaging-protocol <cmd>

    -I --no-imaging
        Don't include the imaging step in the protocol (e.g. so you can provide
        a customized alternative).
"""
    __config__ = [
            DocoptConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.stain'),
    ]
    preset_briefs = appcli.config_attr()

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            Key(DocoptConfig, '<preset>'),
    )
    title = appcli.param(
            Key(PresetConfig),
    )
    light_sensitive = appcli.param(
            Key(PresetConfig),
            default=False,
    )
    prep_buffer = appcli.param(
            Key(PresetConfig),
    )
    prep_volume_mL = appcli.param(
            Key(PresetConfig),
            default=30,
    )
    prep_time_min = appcli.param(
            Key(PresetConfig),
            Key(PresetConfig, 'prep_time_hr', cast=lambda x: x*60),
    )
    prep_microwave = appcli.param(
            Key(PresetConfig),
            default=False,
    )
    prep_steps = appcli.param(
            Key(PresetConfig),
    )
    prep_repeats = appcli.param(
            Key(PresetConfig),
            default=1,
    )
    stain_buffer = appcli.param(
            Key(PresetConfig),
    )
    stain_volume_mL = appcli.param(
            Key(PresetConfig),
            Key(StepwiseConfig, 'default_stain_volume_mL'),
            default=30,
    )
    stain_time_min = appcli.param(
            Key(DocoptConfig, '--time'),
            Key(PresetConfig),
            Key(PresetConfig, 'stain_time_hr', cast=lambda x: x*60),
    )
    stain_microwave = appcli.param(
            Key(PresetConfig),
            default=False,
    )
    stain_steps = appcli.param(
            Key(PresetConfig),
    )
    stain_repeats = appcli.param(
            Key(PresetConfig),
            default=1,
    )
    stain_rinse_repeats = appcli.param(
            Key(PresetConfig),
            default=0,
    )
    destain_buffer = appcli.param(
            Key(PresetConfig),
    )
    destain_volume_mL = appcli.param(
            Key(PresetConfig),
            default=30,
    )
    destain_time_min = appcli.param(
            Key(PresetConfig),
            Key(PresetConfig, 'destain_time_hr', cast=lambda x: x*60),
    )
    destain_microwave = appcli.param(
            Key(PresetConfig),
            default=False,
    )
    destain_steps = appcli.param(
            Key(PresetConfig),
    )
    destain_repeats = appcli.param(
            Key(PresetConfig),
            default=1,
    )
    imaging_cmd = appcli.param(
            Key(DocoptConfig, '--imaging-protocol'),
            Key(DocoptConfig, '--no-imaging', cast=None),
            Key(PresetConfig, 'imaging_protocol'),
            default=None,
    )
    protocol_link = appcli.param(
            Key(PresetConfig),
            default=None,
    )
    footnotes = appcli.param(
            Key(PresetConfig),
            default_factory=list,
    )

    def get_protocol(self):
        p = stepwise.Protocol()

        footnotes = [x] if (x := self.protocol_link) else []
        footnotes += self.footnotes

        s = pl(f"Stain gel with {self.title}{p.add_footnotes(*footnotes)}:")

        if self.light_sensitive:
            #s += "Keep gel in the dark in all following steps."
            s += ul("Keep the stain protected from light.")
            #s += ul("In all following steps, protect the stain from light.")

        s += self._format_stage('prep')
        s += self._format_stage('stain')
        s += self._format_stage('destain')

        p += s

        if self.imaging_cmd:
            p += stepwise.load(self.imaging_cmd)

        return p

    def _format_stage(self, prefix):

        class SkipStage(Exception):
            pass

        def has(attr):
            return hasattr(self, f'{prefix}_{attr}')

        def get(attr, *args):
            return getattr(self, f'{prefix}_{attr}', *args)


        def have_repeats():
            n = get('repeats')
            return n > 1 if isinstance(n, int) else bool(n)

        def format_repeats():
            n = get('repeats')
            if isinstance(n, int):
                n = f'{n}x'
            return f"Repeat {n}:"

        def format_submerge():
            if not has('buffer'): raise SkipStage
            return f"Submerge gel in ≈{get('volume_mL', 30)} mL {get('buffer')}."

        def format_microwave():
            if get('microwave', False):
                return f"Microwave until almost boiling (≈{format_sec(get('microwave_time_s', 45))})."

        def format_incubate():
            return f"Shake gently for {format_min(get('time_min'))}."

        step_templates = get('steps', ['submerge', 'microwave', 'incubate'])
        step_formatters = {
                'submerge': format_submerge,
                'microwave': format_microwave,
                'incubate': format_incubate,
        }

        out = steps = ul()
        for template in step_templates:
            formatter = step_formatters.get(template, lambda: template)
            try:
                steps += formatter()
            except SkipStage:
                return

        if have_repeats():
            out = ul(pl(format_repeats(), steps, br='\n'))

        if n := get('rinse_repeats', False):
            out += f"Rinse {plural(n)://#x }with water."

        return out



if __name__ == '__main__':
    Stain.main()





#!/usr/bin/env python3

from stepwise_mol_bio import _presets as module
from utils import parametrize_via_toml

@parametrize_via_toml('test_presets.toml')
def test_load_preset(presets, key, expected):
    assert module._load_preset(presets, key) == expected

@parametrize_via_toml('test_presets.toml')
def test_get_dotted_key(dict, dotted_key, expected):
    assert module._get_dotted_key(dict, dotted_key) == expected





#!/usr/bin/env python3

import stepwise, appcli, autoprop
from math import sqrt, ceil
from numbers import Real
from inform import plural, indent
from appcli import Key, DocoptConfig
from stepwise import UsageError, StepwiseConfig, PresetConfig, pl, ul, pre
from stepwise_mol_bio import Main, merge_dicts, comma_set, require_reagent
from more_itertools import first_true
from copy import deepcopy
from operator import not_

def temp(x):
    try:
        return float(x)
    except ValueError:
        return x

def time(x):
    return float(x)

@autoprop
class Pcr(Main):
    """\
Amplify a DNA template using polymerase chain reaction (PCR).

Usage:
    pcr <template> <fwd> <rev> (-a <°C>) (-x <sec> | -l <kb>) [options]

Arguments:
    <template>
        The name of the template(s) to amplify.

    <fwd> <rev>
        The names of the forward and reverse primers.

<%! from stepwise_mol_bio import hanging_indent %>\
Options:
    -p --preset <name>              [default: ${app.preset}]
        The default reaction and thermocycler parameters to use.  The following 
        sets of parameters are currently available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

    -l --amplicon-length <bp>
        The length of the amplicon in base pairs (bp).  This can be used to 
        calculate an appropriate extension time.

    -n --num-reactions <int>        [default: ${app.num_reactions}]
        The number of reactions to set up.

    -v --reaction-volume <μL>
        The volume of the PCR reaction.

    -V --template-volume <µL>
        The volume of template to use in the reaction.  This overrides the 
        value specified by the preset.

    -T --template-stock <conc>
        The stock concentration of template to use in the reaction.  This 
        overrides the value specified by the preset, without affecting the 
        volume.  Include a unit, because none is implied.

    -m --master-mix <reagents>      [default: ${','.join(app.master_mix)}]
        Indicate which reagents should be included in the master mix.  The 
        following values are understood:

        dna:      The DNA template.
        fwd:      The forward primer.
        rev:      The reverse primer.
        primers:  Both primers; alias for 'fwd,rev'.

    -M --nothing-in-master-mix
        Don't include anything but water and polymerase in the master mix.  
        This is an alias for: -m ''

    --primer-stock <µM>             [default: ${app.primer_stock_uM}]
        The stock concentration of the primers in µM.  Both primers must have 
        the same concentration.

    -y --num-cycles <n>
        The number of denature/anneal/extend cycles to perform, e.g. 35.

    --initial-denature-temp <°C>
        The temperature of the initial denaturation step in °C, e.g. 95.

    --initial-denature-time <sec>
        The duration of the initial denaturation step in seconds, e.g. 30.

    --denature-temp <°C>
        The temperature of the denaturation step in °C, e.g. 95.

    --denature-time <sec>
        The duration of the denaturation step in seconds, e.g. 10.

    -a --anneal-temp <°C>
        The temperature of the annealing step in °C, e.g. 60.  This is 
        determined by the sequence of the primers.
        
    -g --anneal-temp-gradient <range>
        The range of annealing temperatures that should be tried in a gradient 
        PCR reaction.  The range will be centered at the indicated annealing 
        temperature, and the protocol will indicate the corresponding high and 
        low temperatures.

    --anneal-time <sec>
        The duration of the annealing step in seconds, e.g. 20.

    --extend-temp <°C>
        The temperature of the extension step in °C, e.g. 72.
        
    -x --extend-time <sec>
        The duration of the extension step in seconds.  The rule of thumb is
        30 sec/kb, perhaps longer if you're amplifying a whole plasmid.

    --no-round-extend-time 
        When calculating an extension time from a given amplicon length (e.g. 
        `-l`), do *not* round the result to the nearest 30 sec increment.  Note 
        that when an explicit extension time is given (e.g. `-x`), it is never 
        rounded.

    --final-extend-temp <°C>
        The temperature of the final extension step in °C, e.g. 72.

    --final-extend-time <sec>
        The duration of the annealing step in seconds, e.g. 120.

    --hold-temp <°C>
        The temperature in °C to hold the reaction at after it completes.

    --melt-curve-low-temp <°C>
        The temperature in °C at which to begin recording a melt curve,
        e.g. 65.  This is only relevant for qPCR protocols.

    --melt-curve-high-temp <°C>
        The temperature in °C at which to stop recording a melt curve,
        e.g. 95.  This is only relevant for qPCR protocols.

    --melt-curve-temp-step <°C>
        How much to increment the temperature in °C at each step in the melt 
        curve, e.g. 0.5°C.  This is only relevant for qPCR protocols.

    --melt-curve-time-step <sec>
        The duration in seconds of each step in the melt curve, e.g. 5.  This 
        is only relevant for qPCR protocols.

    --two-step
        Specify that the annealing and extension steps should be combined, e.g. 
        if the primer melting temperatures are very close to the extension 
        temperature or if the amplicon is very short.

    --qpcr
        Specify that this is a qPCR protocol, i.e. that fluorescence should be 
        measured after each thermocyler cycle.
"""

    __config__ = [
            DocoptConfig(usage_getter=lambda self: self.format_usage()),
            PresetConfig(),
            StepwiseConfig('molbio.pcr'),
    ]
    usage = appcli.config_attr()
    preset_briefs = appcli.config_attr()

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            Key(DocoptConfig, '--preset'),
            Key(StepwiseConfig, 'default_preset'),
    )
    template = appcli.param(
            Key(DocoptConfig, '<template>'),
    )
    fwd_primer = appcli.param(
            Key(DocoptConfig, '<fwd>'),
    )
    rev_primer = appcli.param(
            Key(DocoptConfig, '<rev>'),
    )
    primer_stock_uM = appcli.param(
            Key(DocoptConfig, '--primer-stock'),
            Key(PresetConfig, 'primer_stock_uM'),
            Key(StepwiseConfig, 'primer_stock_uM'),
            cast=float,
    )
    amplicon_length_bp = appcli.param(
            Key(DocoptConfig, '--amplicon-length'),
            cast=float,
            default=None,
    )
    num_reactions = appcli.param(
            Key(DocoptConfig, '--num-reactions'),
            cast=lambda x: int(eval(x)),
            default=1,
    )
    reaction_volume_uL = appcli.param(
            Key(DocoptConfig, '--reaction-volume'),
            Key(PresetConfig, 'reaction_volume_uL'),
            cast=lambda x: float(eval(x)),
            default=None,
    )
    template_volume_uL = appcli.param(
            Key(DocoptConfig, '--template-volume'),
            cast=float,
            default=None,
    )
    template_stock = appcli.param(
            Key(DocoptConfig, '--template-stock'),
            default=None,
    )
    master_mix = appcli.param(
            Key(DocoptConfig, '--nothing-in-master-mix', cast=lambda x: set()),
            Key(DocoptConfig, '--master-mix', cast=comma_set),
            Key(StepwiseConfig, 'master_mix', cast=comma_set),
            default={'dna'},
    )
    base_reaction = appcli.param(
            Key(PresetConfig, 'reagents'),
            cast=stepwise.MasterMix.from_text,
    )
    num_cycles = appcli.param(
            Key(DocoptConfig, '--num-cycles'),
            Key(PresetConfig, 'num_cycles'),
            cast=int,
    )
    initial_denature_temp_C = appcli.param(
            Key(DocoptConfig, '--initial-denature-temp'),
            Key(PresetConfig, 'initial_denature_temp_C'),
            cast=temp,
    )
    initial_denature_time_s = appcli.param(
            Key(DocoptConfig, '--initial-denature-time'),
            Key(PresetConfig, 'initial_denature_time_s'),
            cast=time,
    )
    denature_temp_C = appcli.param(
            Key(DocoptConfig, '--denature-temp'),
            Key(PresetConfig, 'denature_temp_C'),
            cast=temp,
    )
    denature_time_s = appcli.param(
            Key(DocoptConfig, '--denature-time'),
            Key(PresetConfig, 'denature_time_s'),
            cast=time,
    )
    anneal_temp_C = appcli.param(
            Key(DocoptConfig, '--anneal-temp'),
            Key(PresetConfig, 'anneal_temp_C'),
            cast=temp,
    )
    anneal_temp_gradient_C = appcli.param(
            Key(DocoptConfig, '--anneal-temp-gradient'),
            Key(PresetConfig, 'anneal_temp_gradient_C'),
            cast=float,
    )
    anneal_time_s = appcli.param(
            Key(DocoptConfig, '--anneal-time'),
            Key(PresetConfig, 'anneal_time_s'),
            cast=time,
    )
    extend_temp_C = appcli.param(
            Key(DocoptConfig, '--extend-temp'),
            Key(PresetConfig, 'extend_temp_C'),
            cast=temp,
    )
    extend_time_s_per_kb = appcli.param(
            Key(PresetConfig, 'extend_time_s_per_kb'),
            cast=float,
    )
    extend_time_func = appcli.param(
            Key(PresetConfig, 'extend_time_func'),
            cast=eval,
    )
    round_extend_time = appcli.param(
            Key(DocoptConfig, '--no-round-extend-time', cast=not_),
            Key(PresetConfig, 'round_extend_time'),
            Key(StepwiseConfig, 'round_extend_time'),
            default=True,
    )
    final_extend_temp_C = appcli.param(
            Key(DocoptConfig, '--final-extend-temp'),
            Key(PresetConfig, 'final_extend_temp_C'),
            cast=temp,
    )
    final_extend_time_s = appcli.param(
            Key(DocoptConfig, '--final-extend-time'),
            Key(PresetConfig, 'final_extend_time_s'),
            cast=time,
    )
    hold_temp_C = appcli.param(
            Key(DocoptConfig, '--hold-temp'),
            Key(PresetConfig, 'hold_temp_C'),
            cast=temp,
            default=None,
    )
    melt_curve_low_temp_C = appcli.param(
            Key(DocoptConfig, '--melt-curve-low-temp'),
            Key(PresetConfig, 'melt_curve_low_temp_C'),
            cast=temp,
    )
    melt_curve_high_temp_C = appcli.param(
            Key(DocoptConfig, '--melt-curve-high-temp'),
            Key(PresetConfig, 'melt_curve_high_temp_C'),
            cast=temp,
    )
    melt_curve_temp_step_C = appcli.param(
            Key(DocoptConfig, '--melt-curve-temp-step'),
            Key(PresetConfig, 'melt_curve_temp_step_C'),
            cast=temp,
    )
    melt_curve_time_step_s = appcli.param(
            Key(DocoptConfig, '--melt-curve-time-step'),
            Key(PresetConfig, 'melt_curve_time_step_s'),
            cast=temp,
    )
    two_step = appcli.param(
            Key(DocoptConfig, '--two-step'),
            Key(PresetConfig, 'two_step'),
            pick=first_true,
            default=False,
    )
    qpcr = appcli.param(
            Key(DocoptConfig, '--qpcr'),
            Key(PresetConfig, 'qpcr'),
            pick=first_true,
            default=False,
    )
    footnote = appcli.param(
            Key(PresetConfig, 'footnote'),
            default=None,
    )

    @appcli.param(
            Key(DocoptConfig, '--extend-time'),
            Key(PresetConfig, 'extend_time_s'),
            cast=time,
            default=None,
    )
    def extend_time_s(self, x):
        if x:
            return x

        bp = self.amplicon_length_bp

        def round(x):
            if not self.round_extend_time: return x
            if x <= 10: return 10
            if x <= 15: return 15
            return 30 * ceil(x / 30)

        if factor := self.extend_time_s_per_kb:
            return round(factor * bp / 1000)

        return round(self.extend_time_func(bp))


    def __init__(self, template, primers):
        self.template = template
        self.primers = primers

    def format_usage(self):
        return self.__doc__

    def get_primers(self):
        return self.fwd_primer, self.rev_primer

    def set_primers(self, primers):
        self.fwd_primer, self.rev_primer = primers

    def get_reaction(self):
        pcr = deepcopy(self.base_reaction)
    
        require_reagent(pcr, 'water')
        require_reagent(pcr, 'template DNA')
        require_reagent(pcr, 'forward primer')
        require_reagent(pcr, 'reverse primer')

        pcr.extra_volume = '10 µL'
        
        if self.num_reactions:
            pcr.num_reactions = self.num_reactions
        if self.reaction_volume_uL:
            pcr.hold_ratios.volume = self.reaction_volume_uL, 'µL'

        pcr['water'].order = 1

        pcr['template DNA'].order = 2
        pcr['template DNA'].name = self.template
        pcr['template DNA'].master_mix = 'dna' in self.master_mix

        if x := self.template_volume_uL:
            pcr['template DNA'].volume = x, 'µL'
        if x := self.template_stock:
            pcr['template DNA'].stock_conc = x

        # Setup the primers.  This is complicated because the primers might 
        # get split into their own mix, if the volumes that would be added 
        # to the PCR reaction are too small.

        primer_mix = None
        primer_abbrev = {
                'forward primer': 'fwd',
                'reverse primer': 'rev',
        }
        use_primer_mix = []

        for p, name in zip(primer_abbrev, self.primers or (None, None)):
            pcr[p].order = 3
            pcr[p].name = name
            pcr[p].hold_conc.stock_conc = self.primer_stock_uM, 'µM'
            pcr[p].master_mix = (
                    primer_abbrev[p] in self.master_mix or
                           'primers' in self.master_mix
            )

            primer_scale = pcr.scale if pcr[p].master_mix else 1
            primer_vol = primer_scale * pcr[p].volume

            if primer_vol < '0.5 µL':
                use_primer_mix.append(p)

        if use_primer_mix:
            pcr['primer mix'].order = 4
            pcr['primer mix'].stock_conc = '10x'
            pcr['primer mix'].volume = pcr.volume / 10
            pcr['primer mix'].master_mix = all(
                    pcr[p].master_mix
                    for p in use_primer_mix
            )

            primer_mix = stepwise.MasterMix()
            primer_mix.volume = pcr.volume

            for p in use_primer_mix:
                primer_mix[p].name = pcr[p].name
                primer_mix[p].stock_conc = pcr[p].stock_conc
                primer_mix[p].volume = pcr[p].volume
                primer_mix[p].hold_stock_conc.conc *= 10
                del pcr[p]

            primer_mix.hold_ratios.volume = '10 µL'

        return pcr, primer_mix

    def get_thermocycler_protocol(self):

        def temp(x):
            if isinstance(x, Real):
                return f'{x:g}°C'
            elif x[-1].isdigit():
                return f'{x}°C'
            else:
                return x

        def time(x):
            try:
                x = int(x)
            except ValueError:
                return x

            if x < 60:
                return f'{x}s'
            elif x % 60:
                return f'{x//60}m{x%60:02}'
            else:
                return f'{x//60} min'

        def has_step(step, params=['temp_C', 'time_s']):
            return all(
                hasattr(self, f'{step}_{param}')
                for param in params
            )

        def step(step):
            if not hasattr(self, f'{step}_temp_gradient_C'):
                t = getattr(self, f'{step}_temp_C')
            else:
                t_mid = getattr(self, f'{step}_temp_C')
                t_range = getattr(self, f'{step}_temp_gradient_C')
                t_low = round(t_mid - t_range / 2)
                t_high = round(t_low + t_range)
                t = f'{t_low}-{t_high}'

            return f"{temp(t)} for {time(getattr(self, f'{step}_time_s'))}"

        three_step = not self.two_step and has_step('extend')

        thermocycler_steps = [
                f"- {step('initial_denature')}",
                f"- Repeat {getattr(self, 'num_cycles')}x:",
                f"  - {step('denature')}",
                f"  - {step('anneal')}",
        ]
        if three_step:
            thermocycler_steps += [
                f"  - {step('extend')}",
            ]

        if self.qpcr:
            thermocycler_steps += [
                f"  - Measure fluorescence",
            ]

        if has_step('final_extend'):
            thermocycler_steps += [
                f"- {step('final_extend')}",
            ]

        if has_step('melt_curve', 'low_temp_C high_temp_C temp_step_C time_step_s'.split()):
            thermocycler_steps += [
                f"- {self.melt_curve_low_temp_C}-{self.melt_curve_high_temp_C}°C in {time(self.melt_curve_time_step_s)} steps of {self.melt_curve_temp_step_C}°C:",
            ]

        if self.qpcr:
            thermocycler_steps += [
                f"  - Measure fluorescence",
            ]

        if self.hold_temp_C:
            thermocycler_steps += [
                f"- {temp(self.hold_temp_C)} hold",
            ]

        return pre('\n'.join(thermocycler_steps))

    def get_protocol(self):
        protocol = stepwise.Protocol()
        pcr, primer_mix = self.reaction
        thermocycler = self.thermocycler_protocol
        footnotes = []

        # Primer mix (if applicable):

        footnotes.append(pl(
                f"For resuspending lyophilized primers:",
                f"{self.primer_stock_uM} µM = {1e3 / self.primer_stock_uM:g} µL/nmol",
                br='\n',
        ))

        if primer_mix:
            protocol += pl(
                    f"Prepare 10x primer mix{protocol.add_footnotes(*footnotes)}:",
                    primer_mix,
            )
            footnotes = []

        # PCR reaction setup:

        if self.footnote:
            # Before the footnote on resuspending the primers, if it hasn't 
            # been added to the protocol already.
            footnotes.insert(0, self.footnote)

        if x := pcr['template DNA'].stock_conc:
            footnotes.append(pl(
                    f"For diluting template DNA to {x}:",
                    f"Dilute 1 µL twice into {sqrt(1000/x.value):.1g}*sqrt([DNA]) µL",
                    br='\n',
            ))

        title = 'qPCR' if self.qpcr else 'PCR'
        instructions = ul()
        protocol += pl(
                f"Setup {plural(pcr.num_reactions):# {title} reaction/s}{protocol.add_footnotes(*footnotes)}:",
                pcr,
                instructions,
        )

        if pcr.volume > '50 µL':
            instructions += f"Split each reaction into {ceil(pcr.volume.value / 50)} tubes."
        if pcr.num_reactions > 1:
            instructions += "Use any extra master mix as a negative control."

        # Thermocycler protocol:

        protocol += pl(
                "Run the following thermocycler protocol:",
                thermocycler,
        )

        protocol.renumber_footnotes()
        return protocol

if __name__ == '__main__':
    Pcr.main()




#!/usr/bin/env python3

import stepwise
import appcli
import autoprop
from inform import plural
from stepwise_mol_bio import UsageError
from stepwise_mol_bio.digest import NebRestrictionEnzymeDatabase
from _assembly import Assembly

@autoprop
class GoldenGate(Assembly):
    """\
Perform a Golden Gate assembly reaction.

Usage:
    golden_gate <backbone> <inserts>... [options]

Arguments:
{FRAGMENT_DOC}

        The only difference between the backbone and the fragments is that the 
        backbone is typically present at half the concentration of the 
        inserts, see --excess-insert.

Options:
{OPTION_DOC}

    -e --enzymes <type_IIS>         [default: {0.enzymes_str}]
        The name(s) of the Type IIS restriction enzyme(s) to use for the 
        reaction.  To use more than one enzyme, enter comma-separated names.  
        The default is to use a single generic name.
"""

    enzymes = appcli.param(
            '--enzymes',
            cast=lambda x: x.split(','),
            default=['BsaI-HFv2'],
    )

    def get_enzymes_str(self):
        # This is why __doc__ should be a Jinja template...
        return ','.join(self.enzymes)

    def get_reaction(self):
        rxn = stepwise.MasterMix()
        rxn.volume = '20 µL'

        rxn['T4 ligase buffer'].volume = '2.0 μL'
        rxn['T4 ligase buffer'].stock_conc = '10x'
        rxn['T4 ligase buffer'].master_mix = True
        rxn['T4 ligase buffer'].order = 2

        enz_uL = 0.5 if len(self.fragments) <= 10 else 1.0

        rxn['T4 DNA ligase'].volume = enz_uL, 'µL'
        rxn['T4 DNA ligase'].stock_conc = '400 U/μL'
        rxn['T4 DNA ligase'].master_mix = True
        rxn['T4 DNA ligase'].order = 3

        enzyme_db = NebRestrictionEnzymeDatabase()

        for enzyme in self.enzymes:
            stock = enzyme_db[enzyme]['concentration'] / 1000
            rxn[enzyme].volume = enz_uL, 'µL'
            rxn[enzyme].stock_conc = stock, 'U/µL'
            rxn[enzyme].master_mix = True
            rxn[enzyme].order = 4

        return self._add_fragments_to_reaction(rxn)

    def get_protocol(self):
        p = stepwise.Protocol()
        n = self.num_reactions

        p += f"""\
Setup the Golden Gate {plural(n):reaction/s} [1]:

{self.reaction}
"""
        if len(self.fragments) == 2:
            p += f"""\
Run the following thermocycler protocol:

- 37°C for 5 min

Or, to maximize the number of transformants:

- 37°C for 60 min
- 60°C for 5 min
"""
        elif len(self.fragments) <= 4:
            p += f"""\
Run the following thermocycler protocol:

- 37°C for 60 min
- 60°C for 5 min
"""
        elif len(self.fragments) <= 10:
            p += f"""\
Run the following thermocycler protocol:

- Repeat 30 times:
  - 37°C for 1 min
  - 16°C for 1 min
- 60°C for 5 min
"""
        else:
            p += f"""\
Run the following thermocycler protocol:

- Repeat 30 times:
  - 37°C for 5 min
  - 16°C for 5 min
- 60°C for 5 min
"""
        p += """\
Transform 2 µL.
"""
        p.footnotes[1] = """\
https://preview.tinyurl.com/yaa5mqz5
"""
        return p

if __name__ == '__main__':
    GoldenGate.main()

# vim: tw=50




#!/usr/bin/env python3

import stepwise

class Presets:

    def __init__(self, presets):
        self.presets = presets

    @classmethod
    def from_config(cls, dotted_key):
        try:
            config = stepwise.load_config()
            config = _get_dotted_key(config, dotted_key)
            return cls(config.data)

        except KeyError:
            return cls({})

    def load(self, key):
        """
        Return a dictionary containing all of the parameters---including 
        inherited parameters---associated with the specified preset.
        """
        return _load_preset(self.presets, key)

    def format_briefs(self, default=''):
        rows = []
        for key in self.presets:
            preset = self.load(key)
            rows.append([
                    f'{key}:',
                    preset.get('brief', default.format(**preset)),
            ])
        return stepwise.tabulate(rows, truncate='-x', max_width=71)

def _load_preset(presets, key):
    from inform import did_you_mean
    from ._utils import ConfigError

    try:
        preset = presets[key]
    except KeyError:
        raise ConfigError(f"no preset {key!r}, did you mean {did_you_mean(str(key), presets)!r}") from None

    if 'inherit' not in preset:
        return preset
    else:
        parent = _load_preset(presets, preset['inherit'])
        merged = {**parent, **preset}; del merged['inherit']
        return merged

def _get_dotted_key(dict, dotted_key):
        for key in dotted_key.split('.'):
            dict = dict[key]
        return dict





#!/usr/bin/env python3

import stepwise, appcli, autoprop
from appcli import DocoptConfig
from stepwise import UsageError, StepwiseConfig, PresetConfig, pl, ul
from stepwise_mol_bio import Main, ConfigError, merge_dicts
from inform import plural

def parse_num_samples(name):
    try:
        return int(name)
    except ValueError:
        return len(name.strip(',').split(','))

def parse_sample_name(name):
    try:
        int(name)
        return None
    except ValueError:
        return name

@autoprop
class Gel(Main):
    """\
Load, run and stain PAGE gels

Usage:
    gel <preset> <samples> [options]

<%! from stepwise_mol_bio import hanging_indent %>\
Arguments:
    <preset>
        What kind of gel to run.  The following presets are available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

    <samples>
        The names of the samples to run, separated by commas.  This can also be 
        a number, which will be taken as the number of samples to run.

Options:
    -p --percent <number>
        The percentage of polyacrylamide/agarose in the gel being run.

    -a --additive <str>
        An extra component to include in the gel itself, e.g. 1x EtBr.

    -b --buffer <str>
        The buffer to run the gel in, e.g. TAE.

    -c --sample-conc <value>
        The concentration of the sample.  This will be used to scale how much 
        sample is mixed with loading buffer, with the goal of mixing the same 
        quantity of material specified in the preset.  In order to use this 
        option, the preset must specify a sample concentration.  The units of 
        that concentration will be used for this concentration.

    -v --sample-volume <µL>
        The volume of sample to mix with loading buffer, in µL.  This does not 
        scale the concentration, and may increase or decrease the amount of 
        sample loaded relative to what's specified in the preset.

    --mix-volume <µL>
        The volume of the sample/loading buffer mix to prepare for each sample.  
        For example, if you want to run two gels, but the preset only makes 
        enough mix for one, use this option to make more.

    --mix-extra <percent>
        How much extra sample/loading buffer mix to make.

    --incubate-temp <°C>
        What temperature to incubate the sample/loading buffer at before 
        loading it onto the gel.  The incubation step will be skipped if 
        neither `--incubate-temp` nor `--incubate-time` are specified (either 
        on the command-line or via the preset).

    --incubate-time <min>
        How long to incubate the sample/loading buffer at the specified 
        temperature before loading it onto the gel.  The incubation step will 
        be skipped if neither `--incubate-temp` nor `--incubate-time` are 
        specified (either on the command-line or via the preset).

    -l --load-volume <µL>
        The volume of the sample/loading buffer mix to load onto the gel.

    --run-volts <V>
        The voltage to run the gel at.

    -r --run-time <min>
        How long to run the gel, in minutes.

    -s --stain <command>
        The name (and arguments) of a protocol describing how to stain the gel.  
        For example, this could be 'gelred' or 'coomassie -f'.
        
    -S --no-stain
        Leave off the staining step of the protocol.
"""

    __config__ = [
            DocoptConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.gel'),
    ]
    preset_briefs = appcli.config_attr()

    presets = appcli.param(
            appcli.Key(StepwiseConfig, 'presets'),
            pick=merge_dicts,
    )
    preset = appcli.param(
            appcli.Key(DocoptConfig, '<preset>'),
    )
    title = appcli.param(
            appcli.Key(PresetConfig, 'title'),
            default='electrophoresis',
    )
    num_samples = appcli.param(
            appcli.Key(DocoptConfig, '<samples>', cast=parse_num_samples),
            ignore=None,
            default=1,
    )
    sample_name = appcli.param(
            appcli.Key(DocoptConfig, '<samples>', cast=parse_sample_name),
            default=None,
    )
    sample_mix = appcli.param(
            appcli.Key(PresetConfig, 'sample_mix'),
            default=None,
    )
    sample_conc = appcli.param(
            appcli.Key(DocoptConfig, '--sample-conc'),
            appcli.Key(PresetConfig, 'sample_conc'),
            cast=float,
            default=None,
    )
    sample_volume_uL = appcli.param(
            appcli.Key(DocoptConfig, '--sample-volume'),
            appcli.Key(PresetConfig, 'sample_volume_uL'),
            cast=float,
            default=None,
    )
    ladder_name = appcli.param(
            appcli.Key(PresetConfig, 'ladder'),
            default=None,
    )
    ladder_volume_uL = appcli.param(
            appcli.Key(PresetConfig, 'ladder_volume_uL'),
            default=None,
    )
    mix_volume_uL = appcli.param(
            appcli.Key(DocoptConfig, '--mix-volume'),
            appcli.Key(PresetConfig, 'mix_volume_uL'),
            cast=float,
            default=None,
    )
    mix_extra_percent = appcli.param(
            appcli.Key(DocoptConfig, '--mix-extra'),
            appcli.Key(PresetConfig, 'mix_extra_percent'),
            cast=float,
            default=50,
    )
    incubate_temp_C = appcli.param(
            appcli.Key(DocoptConfig, '--incubate-temp'),
            appcli.Key(PresetConfig, 'incubate_temp_C'),
            cast=float,
    )
    incubate_time_min = appcli.param(
            appcli.Key(DocoptConfig, '--incubate-time'),
            appcli.Key(PresetConfig, 'incubate_time_min'),
            cast=int,
    )
    gel_type = appcli.param(
            appcli.Key(PresetConfig, 'gel_type'),
    )
    gel_percent = appcli.param(
            appcli.Key(DocoptConfig, '--percent'),
            appcli.Key(PresetConfig, 'gel_percent'),
    )
    gel_additive = appcli.param(
            appcli.Key(DocoptConfig, '--additive'),
            appcli.Key(PresetConfig, 'gel_additive'),
            default=None,
    )
    gel_buffer = appcli.param(
            appcli.Key(DocoptConfig, '--buffer'),
            appcli.Key(PresetConfig, 'gel_buffer'),
    )
    load_volume_uL = appcli.param(
            appcli.Key(DocoptConfig, '--load-volume'),
            appcli.Key(PresetConfig, 'load_volume_uL'),
            cast=float,
    )
    run_volts = appcli.param(
            appcli.Key(DocoptConfig, '--run-volts'),
            appcli.Key(PresetConfig, 'run_volts'),
            cast=float,
    )
    run_time_min = appcli.param(
            appcli.Key(DocoptConfig, '--run-time'),
            appcli.Key(PresetConfig, 'run_time_min'),
            cast=int,
    )
    stain = appcli.param(
            appcli.Key(DocoptConfig, '--stain'),
            appcli.Key(DocoptConfig, '--no-stain', cast=lambda x: None),
            appcli.Key(PresetConfig, 'stain'),
            default=None,
    )

    def __init__(self, preset, num_samples=None):
        self.preset = preset
        self.num_samples = num_samples

    def get_protocol(self):
        p = stepwise.Protocol()

        def both_or_neither(key1, key2):
            has_key1 = has_key2 = True

            try: value1 = getattr(self, key1)
            except AttributeError: has_key1 = False

            try: value2 = getattr(self, key2)
            except AttributeError: has_key2 = False

            if has_key1 and not has_key2:
                raise ConfigError(f"specified {key1!r} but not {key2!r}")
            if has_key2 and not has_key1:
                raise ConfigError(f"specified {key2!r} but not {key1!r}")

            if has_key1 and has_key2:
                return value1, value2
            else:
                return False

        if self.sample_mix:
            mix = stepwise.MasterMix.from_text(self.sample_mix)
            mix.num_reactions = self.num_samples
            mix.extra_percent = self.mix_extra_percent
            mix['sample'].name = self.sample_name

            if x := self.sample_conc:
                stock_conc = mix['sample'].stock_conc
                if stock_conc is None:
                    raise ConfigError(f"can't change sample stock concentration, no initial concentration specified.")
                mix['sample'].hold_conc.stock_conc = x, stock_conc.unit

            if x := self.sample_volume_uL:
                mix['sample'].volume = x, 'µL'

            if x := self.mix_volume_uL:
                mix.hold_ratios.volume = x, 'µL'

            mix.fix_volumes('sample')

            p += pl(
                    f"Prepare {plural(self.num_samples):# sample/s} for {self.title}:",
                    mix,
            )
            if x := both_or_neither('incubate_temp_C', 'incubate_time_min'):
                temp_C, time_min = x
                p.steps[-1] += ul(
                        f"Incubate at {temp_C:g}°C for {time_min:g} min."
                )
            
        additive = f" with {x}" if (x := self.gel_additive) else ""
        p += pl("Run a gel:", ul(
            f"Gel: {self.gel_percent}% {self.gel_type}{additive}",
            f"Buffer: {self.gel_buffer}",
            f"Ladder: {self.ladder_volume_uL:g} µL {self.ladder_name}"
                if self.ladder_name else None,
            f"Load: {self.load_volume_uL:g} µL",
            f"Run: {self.run_volts:g}V for {self.run_time_min:g} min",
        ))

        if x := self.stain:
            p += stepwise.load(x)

        return p

if __name__ == '__main__':
    Gel.main()



#!/usr/bin/env python3

import sys
import appcli
from appdirs import AppDirs
from inform import Error, format_range
from pathlib import Path

app_dirs = AppDirs("stepwise_mol_bio")

class Main(appcli.App):
    usage_io = sys.stderr

    @classmethod
    def main(cls):
        self = cls.from_params()
        self.load()
        
        try:
            self.protocol.print()
        except Error as err:
            err.report()



class StepwiseMolBioError(Error):
    pass

class ConfigError(StepwiseMolBioError):
    pass

class UsageError(StepwiseMolBioError):
    pass

def try_except(expr, exc, failure, success=None):
    try:
        x = expr()
    except exc:
        return failure()
    else:
        return success() if success else x

def hanging_indent(text, prefix):
    from textwrap import indent
    if isinstance(prefix, int):
        prefix = ' ' * prefix
    return indent(text, prefix)[len(prefix):]

def merge_dicts(dicts):
    result = {}
    for dict in reversed(list(dicts)):
        result.update(dict)
    return result

def comma_list(x):
    return [x.strip() for x in x.split(',')]

def comma_set(x):
    return {x.strip() for x in x.split(',')}

def require_reagent(rxn, reagent):
    if reagent not in rxn:
        raise UsageError(f"reagent table missing {reagent!r}")


def format_sec(x):
    if x < 60:
        return f'{x}s'

    min = x // 60
    sec = x % 60

    return f'{min}m{f"{sec:02}" if sec else ""}'

def format_min(x):
    if x < 60:
        return f'{x}m'

    hr = x // 60
    min = x % 60

    return f'{hr}h{f"{sec:02}" if min else ""}'




#!/usr/bin/env python3

import stepwise, appcli, autoprop
from stepwise_mol_bio import Main

@autoprop
class Aliquot(Main):
    """\
Make aliquots

Usage:
    aliquot <volume> [<conc>]

Arguments:
    <volume>
        The volume of each individual aliquot.  No unit is implied, so you 
        should specify one.

    <conc>
        The concentration of the aliquots, if this is not made clear in 
        previous steps.  No unit is implied, so you should specify one.
"""
    __config__ = [
            appcli.DocoptConfig(),
    ]
    volume = appcli.param('<volume>')
    conc = appcli.param('<conc>', default=None, ignore=None)

    def __init__(self, volume, conc=None):
        self.volume = volume
        self.conc = conc

    def get_protocol(self):
        p = stepwise.Protocol()

        if self.conc:
            p += f"Make {self.volume}, {self.conc} aliquots."
        else:
            p += f"Make {self.volume} aliquots."

        return p

if __name__ == '__main__':
    Aliquot.main()



#!/usr/bin/env python3

import stepwise, appcli, autoprop
from appcli import Key, DocoptConfig
from stepwise import paragraph_list, unordered_list
from stepwise_mol_bio import Main, UsageError
from stepwise_mol_bio.gels.gel import Gel
from operator import not_
from inform import plural

@autoprop
class Ivt(Main):
    """\
Display a protocol for synthesizing in vitro transcribed (IVT) RNA using the
NEB HiScribe kit (E2040).

Usage:
    ivt <templates>... [options]

Arguments:
    <templates>
        The names to the DNA templates to transcribe.

Options:
    -d --dna-ng-uL CONC  [default: ${app.dna_ng_uL}]
        The stock concentration of the template DNA (in ng/µL).

    -D --dna-ng AMOUNT    [default: ${app.dna_ng}]
        How much template DNA to use (in ng).  NEB recommends 1 µg.  Lower 
        amounts will give proportionately lower yield, but will otherwise work 
        fine.  If the template is not concentrated enough to reach the given 
        quantity, the reaction will just contain as much DNA as possible 
        instead.

    -V --dna-uL VOLUME
        The volume of DNA to add to the reaction (in µL).  By default, this 
        will be inferred from --dna-ng and --dna-ng-uL.

    -s --short
        Indicate that all of the templates are shorter than 300 bp.  This 
        allows the reactions to be setup with less material, so that more 
        reactions can be performed with a single kit.

    -i --incubate HOURS
        How long to incubate the transcription reaction.  The default is 2h for 
        long transcripts and 4h for short transcripts (--short).

    -x --extra PERCENT      [default: ${app.extra_percent}]
        How much extra master mix to create.

    -R --no-rntp-mix
        Indicate that each you're not using a rNTP mix and that you need to add 
        each rNTP individually to the reaction.

    -c --cleanup METHOD     [default: ${app.cleanup}]
        Choose the method for removing free nucleotides from the RNA:

        zymo: Zymo spin kits.
        ammonium: Ammonium acetate precipitation.

    -C --no-cleanup
        Don't include a cleanup step.

    -G --no-gel
        Don't include the gel electrophoresis step.

Template Preparation:
    The following information is taken directly from NEB:
    https://www.neb.com/protocols/0001/01/01/dna-template-preparation-e2040

    Plasmid Templates:
        Completely linearized plasmid template of highest purity is critical 
        for successful use of the HiScribe T7 High Yield RNA Synthesis Kit. 
        Quality of the template DNA affects transcription yield and the 
        integrity of RNA synthesized. The highest transcription yield is 
        achieved with the highest purity template. Plasmid purified by many 
        laboratory methods can be successfully used, provided it contains 
        mostly supercoiled form, and is free from contaminating RNase, protein, 
        RNA and salts.

        To produce RNA transcript of a defined length, plasmid DNA must be 
        completely linearized with a restriction enzyme downstream of the 
        insert to be transcribed. Circular plasmid templates will generate long 
        heterogeneous RNA transcripts in higher quantities because of high 
        processivity of T7 RNA polymerase. NEB has a large selection of 
        restriction enzymes; we recommend selecting restriction enzymes that 
        generate blunt ends or 5´-overhangs.

        After linearization, we recommend purifying the template DNA by 
        phenol/chloroform extraction:

        1. Extract DNA with an equal volume of 1:1 phenol/chloroform mixture, 
           repeat if necessary.
        2. Extract twice with an equal volume of chloroform to remove residual 
           phenol.
        3. Precipitate the DNA by adding 1/10th volume of 3 M sodium acetate, 
           pH 5.2, and two volumes of ethanol. Incubate at -20°C for at least 
           30 minutes.
        4. Pellet the DNA in a microcentrifuge for 15 minutes at top speed. 
           Carefully remove the supernatant.
        5. Rinse the pellet by adding 500 μl of 70% ethanol and centrifuging 
           for 15 minutes at top speed. Carefully remove the supernatant.
        6. Air dry the pellet and resuspend it in nuclease-free water at a 
           concentration of 0.5-1 μg/μl.

    PCR Templates:

        PCR products containing T7 RNA Polymerase promoter in the correct 
        orientation can be transcribed. Though PCR mixture can be used 
        directly, better yields will be obtained with purified PCR products. 
        PCR products can be purified according to the protocol for plasmid 
        restriction digests above, or by using commercially available spin 
        columns (we recommend Monarch PCR & DNA Cleanup Kit, NEB #T1030). PCR 
        products should be examined on an agarose gel to estimate concentration 
        and to confirm amplicon size prior to its use as a template in the 
        HiScribe T7 High Yield RNA Synthesis Kit. Depending on the PCR 
        products, 0.1–0.5 μg of PCR fragments can be used in a 20 μl in vitro 
        transcription reaction.

    Synthetic DNA Oligonucleotides:

        Synthetic DNA Oligonucleotides which are either entirely 
        double-stranded or mostly single-stranded with a double-stranded T7 
        promoter sequence can be used in the HiScribe T7 High Yield RNA 
        Synthesis Kit. In general, the yields are relatively low and also 
        variable depending upon the sequence, purity and preparation of the 
        synthetic oligonucleotides.
"""
    __config__ = [
            DocoptConfig(),
    ]

    templates = appcli.param(
            Key(DocoptConfig, '<templates>'),
    )
    _dna_uL = appcli.param(
            Key(DocoptConfig, '--dna-uL'),
            cast=float,
            default=None,
    )
    _dna_ng = appcli.param(
            Key(DocoptConfig, '--dna-ng'),
            cast=float,
            default=1000,
    )
    dna_ng_uL = appcli.param(
            Key(DocoptConfig, '--dna-ng-uL'),
            cast=float,
            default=500,
    )
    cleanup = appcli.param(
            Key(DocoptConfig, '--no-cleanup', cast=lambda x: 'none'),
            Key(DocoptConfig, '--cleanup'),
            default='zymo',
    )
    short = appcli.param(
            Key(DocoptConfig, '--short'),
            default=False,
    )
    incubation_h = appcli.param(
            Key(DocoptConfig, '--incubate'),
            default=None,
    )
    rntp_mix = appcli.param(
            Key(DocoptConfig, '--no-rntp-mix', cast=not_),
            default=True,
    )
    gel = appcli.param(
            Key(DocoptConfig, '--no-gel', cast=not_),
            default=True,
    )
    extra_percent = appcli.param(
            Key(DocoptConfig, '--extra-percent'),
            cast=float,
            default=10,
    )

    def __init__(self, templates):
        self.templates = templates

    def get_reaction(self):
        ivt = stepwise.MasterMix("""\
                Reagent              Product #  Stock Conc      Volume  MM?
                ===================  =========  ==========  ==========  ===
                nuclease-free water                         to 20.0 μL  yes
                reaction buffer                        10x      2.0 μL  yes
                rATP                                100 mM      2.0 μL  yes
                rCTP                                100 mM      2.0 μL  yes
                rGTP                                100 mM      2.0 μL  yes
                rUTP                                100 mM      2.0 μL  yes
                HiScribe T7          NEB E2040         10x      2.0 μL  yes
                DNA template                     500 ng/μL      2.0 μL
        """)
        ivt.extra_percent = self.extra_percent
        ivt.num_reactions = len(self.templates)

        ivt['nuclease-free water'].order = 1
        ivt['reaction buffer'].order = 2
        ivt['rATP'].order = 3
        ivt['rCTP'].order = 3
        ivt['rGTP'].order = 3
        ivt['rUTP'].order = 3
        ivt['HiScribe T7'].order = 4
        ivt['DNA template'].order = 5

        if self.short:
            ivt['reaction buffer'].volume = '1.5 µL'
            ivt['rATP'].volume = '1.5 µL'
            ivt['rCTP'].volume = '1.5 µL'
            ivt['rGTP'].volume = '1.5 µL'
            ivt['rUTP'].volume = '1.5 µL'

        if self.rntp_mix:
            ivt['rNTP mix'].volume = 4 * ivt['rATP'].volume
            ivt['rNTP mix'].stock_conc = '100 mM'
            ivt['rNTP mix'].order = 3
            del ivt['rATP']
            del ivt['rCTP']
            del ivt['rGTP']
            del ivt['rUTP']

        ivt['DNA template'].name = ','.join(self.templates)
        ivt['DNA template'].stock_conc = self.dna_ng_uL, 'ng/µL'

        if self.dna_uL:
            ivt['DNA template'].volume = self.dna_uL, 'µL'
        else:
            ivt['DNA template'].volume = self.dna_ng / self.dna_ng_uL, 'µL'
        
        ivt.fix_volumes('DNA template', 'nuclease-free water')
        return ivt

    def get_protocol(self):
        p = stepwise.Protocol()

        ## Clean your bench
        p += """\
Wipe down your bench and anything you'll touch 
(pipets, racks, pens, etc.) with RNaseZap.
"""
        ## In vitro transcription
        ivt = self.reaction
        n = plural(ivt.num_reactions)
        p += paragraph_list(
                f"Setup {n:# in vitro transcription reaction/s} [1,2]:",
                ivt,
                unordered_list(
                    "Mix reagents in the order given.",
                    "Ok to handle at room temperature.",
                ),
        )
        p.footnotes[1] = """\
https://tinyurl.com/y4a2j8w5
"""
        p.footnotes[2] = """\
I've found that T7 kits which have been in the 
freezer for more than ≈4 weeks seem to produce 
more degraded RNA.
"""
        t = self.incubation_h or (4 if self.short else 2)
        p += f"""\
Incubate at 37°C for {plural(t):# hour/s} [3].
"""
        p.footnotes[3] = """\
Use a thermocycler to prevent evaporation.
"""
        ## Purify product
        if self.cleanup == 'zymo':
            p += """\
Remove unincorporated ribonucleotides using Zymo 
RNA Clean & Concentrator 25 spin columns.
"""
        elif self.cleanup == 'ammonium':
            if self.short:
                raise UsageError("ammonium acetate precipitation cannot be used for short transcripts (<100 bp)")
            protocol += """\
Remove unincorporated ribonucleotides using
ammonium acetate precipitation:

- Add 1 volume (20 μL) 5M ammonium acetate to 
  each reaction.
- Incubate on ice for 15 min.
- Centrifuge at >10,000g for 15 min at 4°C.
- Wash pellet with 70% ethanol.
- Dissolve pellet in 20μL nuclease-free water.
"""
        elif self.cleanup == 'none':
            return p
        else:
            raise UsageError(f"unknown RNA clean-up method: '{self.cleanup}'")

#        ## Nanodrop concentration
#        p += """\
#Nanodrop to determine the RNA concentration.
#"""
#        ## Aliquot
#        p += """\
#Dilute (if desired) enough RNA to make several 
#10 μM aliquots and to run a gel.  Keep any left- 
#over RNA undiluted.  Flash-freeze in liquid N₂
#and store at -80°C.
#"""
#        ## Gel electrophoresis
#        if self.gel:
#            p += Gel('urea', len(self.templates)).protocol
#
        return p

    def get_dna_uL(self):
        return self._dna_uL

    def get_dna_ng(self):
        return self._dna_ng

    def set_dna_uL(self, dna_uL):
        self._dna_uL = dna_uL
        self._dna_ng = self.dna_ng_uL / dna_uL

    def set_dna_ng(self, dna_ng):
        self._dna_ng = dna_ng
        self._dna_uL = dna_ng / self.dna_ng_uL

    @appcli.on_load
    def _load_dna_quantity(self):
        if self._dna_uL:
            self.set_dna_uL(self._dna_uL)
        else:
            self.set_dna_ng(self._dna_ng)

if __name__ == '__main__':
    Ivt.main()

# vim: tw=50



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['stepwise_mol_bio', 'stepwise_mol_bio.gels']

package_data = \
{'': ['*'], 'stepwise_mol_bio': ['layouts/*']}

install_requires = \
['appdirs',
 'autoprop',
 'configurator',
 'docopt',
 'inform>=1.21',
 'numpy',
 'pytest',
 'requests',
 'stepwise',
 'voluptuous']

extras_require = \
{'docs': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'tests': ['pytest', 'pytest-cov', 'coveralls', 'toml']}

entry_points = \
{'po4.syntheses': ['PCR = stepwise_mol_bio.pcr:Po4Synthesis'],
 'stepwise.protocols': ['molbio = stepwise_mol_bio:Plugin']}

setup(name='stepwise_mol_bio',
      version='1.12.0',
      description='Protocols relating to molecular biology, e.g. PCR.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/stepwise_mol_bio',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='~=3.8',
     )



#!/usr/bin/env python3

import stepwise
import appcli
import autoprop
import textwrap
from inform import plural
from fractions import Fraction
from operator import not_
from appcli import Key, DocoptConfig
from stepwise import StepwiseConfig, PresetConfig, pl, ul, pre
from stepwise_mol_bio import Main

def by_solvent(obj, x):
    if isinstance(x, dict):
        return x[obj.solvent.lower()]
    else:
        return x

@autoprop
class EthanolPrecipitation(Main):
    """\
Purify and concentrate nucleic acids by ethanol precipitation.

This protocol is primarily based on [Li2020].

Usage:
    ethanol_precipitation [<names>...] [options]

Arguments:
    <names>
        The names of the constructs to precipitate.

Options:
    -p --preset <name>                    [default: ${app.preset}]
        There are four versions of the protocol, each optimized for a different 
        nucleic acid species.  Use this option to specify which version to use.  
        The names are case-insensitive:

        plasmid:
            Optimized with 10 kb circular plasmid.  This protocol is probably 
            also most appropriate for linear molecules of comparable size (e.g. 
            restriction digested plasmids).

        pcr:
            Optimized with 150 bp linear, doubled-stranded DNA.

        primer:
            Optimized with 20 nt single-stranded DNA.

        microrna:
            Optimized with 20 nt single-stranded RNA.

    -s --solvent <name>
        The organic solvent to use for the precipitation.  The names are 
        case-insensitive.

        etoh: Ethanol
            - Gives higher yield than isopropanol for short RNA/DNA, and 
              comparable yield for longer DNA [Li2020].
            - Evaporates more easily after the precipitation.

        iproh: Isopropanol
            - Has been said to work better than ethanol for dilute samples, 
              although this was not tested by [Li2020].
            - Requires less volume, which may be beneficial when working with 
              large volumes.
            - Better at dissolving (and therefore removing) protein and 
              polysaccharide contaminants.
            - Precipitates more salt, resulting in higher salt contamination.

    -a --cation <name>
        The cation to use for the precipitation.  This is automatically 
        determined by the protocol, but you can specify a different choice 
        (e.g. based on what you have on hand).  The names are case-insensitive:

        na: ${app.cations['na']['conc']} ${app.cations['na']['name']}
        mg: ${app.cations['mg']['conc']} ${app.cations['mg']['name']}

        Other cations were tested in [Li2020], but either NaAc or MgCl₂ was the 
        best in every condition.

    -c --carrier <name>
        The carrier, or coprecipitator, to add to the reaction.  This is 
        automatically determined by the protocol, but you can specify a 
        different choice (e.g. based on what you have on hand).  The names are 
        case-insensitive:

        lpa: ${app.carriers['lpa']['name']}
            Not known to interfere with any downstream application.  Not 
            derived from a biological source, so very unlikely to have any 
            nucleic acid contamination.

        glycogen:
            Mostly inert, but may interfere with protein/DNA interactions 
            [Gaillard1990] and reverse transcription (at concentrations 
            >2 mg/mL).  Derived from biological source, so may contain trace 
            contaminating nucleic acids.  You can purchase glycogen crosslinked 
            to a blue dye, which makes the pellet even easier to see.

        trna: ${app.carriers['trna']['name']}
            Interferes with the quantification of the nucleic acid by Nanodrop, 
            which is problematic for many applications.

    -b --buffer <name>                      [default: ${app.buffer}]
        The aqueous buffer to resuspend the precipitated nucleic acid in.

    -v --buffer-volume <µL>
        The volume of resuspension buffer to use, in µL.

    -I --no-incubation
        Exclude the incubation step.

    -W --no-wash
        Exclude the wash step.
            
References:
    Li Y et al.  A systematic investigation of key factors of nucleic acid 
    precipitation toward optimized DNA/RNA isolation.  BioTechniques 68, 
    191–199 (2020).

    Gaillard C, Strauss F.  Ethanol precipitation of DNA with linear 
    polyacrylamide as carrier.  Nucleic Acids Res. 18(2), 378 (1990).

    Sambrook J & Russell DW.  Standard ethanol precipitation of DNA in 
    microcentrifuge tubes.  Cold Spring Harb Protoc (2006).
"""

    __config__ = [
            DocoptConfig(),
            PresetConfig(),
            StepwiseConfig('molbio.ethanol_precipitation'),
    ]

    presets = {
            'plasmid': {
                'solvent': 'etoh',
                'solvent_volume': {
                    'etoh': 3,
                    'iproh': 1,
                },

                'cation': {
                    'etoh': 'na',
                    'iproh': 'na',
                },

                'carrier': {
                    'etoh': 'lpa',
                    'iproh': 'lpa',
                },

                'incubation_time': None,
                'incubation_temp_C': None,

                'centrifugation_time_min': 60,
                'centrifugation_temp_C': 4,
                'centrifugation_speed': '>7500g',
            },
            'pcr': {
                'solvent': 'etoh',
                'solvent_volume': {
                    'etoh': 2,
                    'iproh': Fraction(3,4),
                },

                'cation': {
                    'etoh': 'mg',
                    'iproh': 'mg',
                },

                'carrier': {
                    'etoh': 'glycogen',
                    'iproh': 'lpa',
                },

                'incubation_time': 'overnight',
                'incubation_temp_C': -20,

                'centrifugation_time_min': 60,
                'centrifugation_temp_C': 4,
                'centrifugation_speed': '>7500g'
            },
            'primer': {
                'solvent': 'etoh',
                'solvent_volume': {
                    'etoh': 4,
                    'iproh': 1,
                },

                'cation': {
                    'etoh': 'na',
                    'iproh': 'mg',
                },

                'carrier': {
                    'etoh': 'glycogen',
                    'iproh': 'glycogen',
                },

                'incubation_time': 'overnight',
                'incubation_temp_C': 4,

                'centrifugation_time_min': 60,
                'centrifugation_temp_C': 4,
                'centrifugation_speed': '>18000g',
            },
            'microrna': {
                'solvent': 'etoh',
                'solvent_volume': {
                    'etoh': 4,
                    'iproh': Fraction(3,4),
                },

                'cation': {
                    'etoh': 'mg',
                    'iproh': 'na',
                },

                'carrier': {
                    'etoh': 'glycogen',
                    'iproh': 'lpa',
                },

                'incubation_time': 'overnight',
                'incubation_temp_C': -20,

                'centrifugation_time_min': 60,
                'centrifugation_temp_C': 4,
                'centrifugation_speed': '>21000',
            },
    }
    solvents = {
            'etoh': {
                'name': '100% ethanol',
            },
            'iproh': {
                'name': 'isopropanol',
            },
    }
    carriers = {
            'trna': {
                'name': "yeast tRNA",
                'conc': "20 ng/µL",
            },
            'glycogen': {
                'name': "glycogen",
                'conc': "50 ng/µL",
            },
            'lpa': {
                'name': "linear polyacrylamide (LPA)",
                'conc': "20 ng/µL",
            },
    }
    cations = {
            'na': {
                'name': "sodium acetate, pH=5.2",
                'conc': "300 mM",
            },
            'mg': {
                'name': "magnesium chloride (MgCl₂)",
                'conc': "10 mM",
            },
    }

    preset = appcli.param(
            Key(DocoptConfig, '--preset'),
            Key(StepwiseConfig, 'preset'),
            ignore=None,
    )
    names = appcli.param(
            Key(DocoptConfig, '<names>'),
            default=None,
    )
    solvent = appcli.param(
            Key(DocoptConfig, '--solvent'),
            Key(PresetConfig, 'solvent'),
    )
    solvent_volume = appcli.param(
            Key(PresetConfig, 'solvent_volume'),
            get=by_solvent,
    )
    buffer = appcli.param(
            Key(DocoptConfig, '--buffer'),
            default='water',
    )
    buffer_volume_uL = appcli.param(
            Key(DocoptConfig, '--buffer-volume'),
            default=None,
    )
    cation = appcli.param(
            Key(DocoptConfig, '--cation'),
            Key(PresetConfig, 'cation'),
            get=by_solvent,
    )
    carrier = appcli.param(
            Key(DocoptConfig, '--carrier'),
            Key(PresetConfig, 'carrier'),
            get=by_solvent,
    )
    incubation = appcli.param(
            Key(DocoptConfig, '--no-incubation', cast=not_),
            default=True,
    )
    incubation_time = appcli.param(
            Key(PresetConfig, 'incubation_time'),
    )
    incubation_temp_C = appcli.param(
            Key(PresetConfig, 'incubation_temp_C'),
    )
    wash = appcli.param(
            Key(DocoptConfig, '--no-wash', cast=not_),
            default=True,
    )
    centrifugation_time_min = appcli.param(
            Key(PresetConfig, 'centrifugation_time_min'),
    )
    centrifugation_temp_C = appcli.param(
            Key(PresetConfig, 'centrifugation_temp_C'),
    )
    centrifugation_speed = appcli.param(
            Key(PresetConfig, 'centrifugation_speed'),
    )

    def __init__(self, preset=None):
        self.preset = preset

    def get_protocol(self):
        p = stepwise.Protocol()
        s = ul()

        if self.names:
            p += pl(f"Purify {','.join(self.names)} by ethanol precipitation [1,2]:", s)
        else:
            p += pl("Perform an ethanol precipitation [1,2]:", s)

        s += f"""\
                Add {self.cation_name} to {self.cation_conc}."""
        s += f"""\
                Add {self.carrier_name} to {self.carrier_conc}."""
        s += f"""\
                Add {plural(self.solvent_volume):# volume/s} 
                {self.solvent_name} and mix well."""
        s += f"""\
                If necessary, divide the sample between microfuge tubes
                such that none holds more than 400 µL."""

        if self.incubation and (t := self.incubation_time):
            incubation_time = "overnight" if t == 'overnight' else f"for {t}"
            s += f"""\
                    Incubate at {self.incubation_temp_C}°C {incubation_time} 
                    [3]."""

        s += f"""\
                Centrifuge {self.centrifugation_speed}, 
                {self.centrifugation_time_min} min, 
                {self.centrifugation_temp_C}°C.  Remove the supernatant, 
                but save it in case the precipitation needs to be repeated."""

        if self.wash:
            s += f"""\
                    Add 800 µL recently-prepared 70% ethanol [4]."""
            s += f"""\
                    Centrifuge {self.centrifugation_speed}, 2 min, 
                    {self.centrifugation_temp_C}°C.  Discard supernatant."""
        s += f"""\
                Centrifuge {self.centrifugation_speed}, 30 s, 
                {self.centrifugation_temp_C}°C.  Discard any remaining 
                supernatant.
        """
        s += f"""\
                Leave the tube open at room temperature until ethanol has 
                evaporated [5]."""

        s += f"""\
                Resuspend the pellet in {f'{self.buffer_volume_uL} µL' if 
                self.buffer_volume_uL else 'any volume'} of {self.buffer} 
                [6]."""

        p.footnotes[1] = pre(textwrap.dedent("""\
                Li2020: 10.2144/btn-2019-0109
                Sambrook2006: 10.1101/pdb.prot4456"""
        ))
        p.footnotes[2] = """\
                This protocol was optimized for 100 ng/µL nucleic acid.  If 
                your sample is substantially more dilute, it may be necessary 
                to compensate by increasing the incubation time, the 
                centrifugation time, or the centrifugation speed.
        """
        p.footnotes[3] = """\
                DNA can be stored indefinitely in ethanolic solutions at either 
                0°C or −20°C.
        """
        p.footnotes[4] = """\
                Ethanol evaporates more quickly than water, so a solution that 
                was 70% ethanol several months ago may be significantly more 
                aqueous now.  If you are unsure, 100 µL of 70% EtOH should 
                weigh 88.6 mg.
        """
        p.footnotes[5] = """\
                Do not dry pellets of nucleic acid in a lyophilizer, as this 
                causes denaturation of small (<400-nucleotide) fragments of DNA 
                and greatly reduces the recovery of larger fragments of DNA. 

                If necessary, the open tube containing the redissolved DNA can 
                be incubated for 2-3 minutes at 45°C in a heating block to 
                allow any traces of ethanol to evaporate.
        """
        p.footnotes[6] = """\
                Up to 50% of the DNA is smeared on the wall of the tube. To 
                recover all of the DNA, push a bead of fluid backward and 
                forward over the appropriate quadrant of wall with a pipette 
                tip.
        """

        p.prune_footnotes()
        return p

    def get_solvent_name(self):
        return self.solvents[self.solvent]['name']

    def get_cation_name(self):
        return self.cations[self.cation]['name']

    def get_cation_conc(self):
        return self.cations[self.cation]['conc']

    def get_carrier_name(self):
        return self.carriers[self.carrier]['name']

    def get_carrier_conc(self):
        return self.carriers[self.carrier]['conc']

if __name__ == '__main__':
    EthanolPrecipitation.main()
